'''
    EEGNet-8,2
    https://iopscience.iop.org/article/10.1088/1741-2552/aace8c/meta
'''


import torch
from torch import Tensor, nn
from torchinfo import summary


class Conv2dWithNorm(nn.Conv2d):
    
    def __init__(self, *args, maxNorm, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.maxNorm = maxNorm
        
    def forward(self, input: Tensor) -> Tensor:
        self.weight.data = torch.renorm(
            self.weight.data, 2, 0, self.maxNorm
        )
        return super().forward(input)


class LinearWithNorm(nn.Linear):
    
    def __init__(self, *args, maxNorm, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.maxNorm = maxNorm
    
    def forward(self, input: Tensor) -> Tensor:
        self.weight.data = torch.renorm(
            self.weight.data, 2, 0, self.maxNorm
        )
        return super().forward(input)


class EEGNet(nn.Module):
    
    def __init__(self, nCh, nTime, kernel=125, F1=8, D=2, F2=16,
                 p=0.25, cls=4, bias=False) -> None:
        super().__init__()
        
        self.filter = nn.Sequential(
            nn.Conv2d(1, F1, (1, kernel), padding='same', bias=bias),
            nn.BatchNorm2d(F1)
        )
        
        self.depthwiseConv = nn.Sequential(
            Conv2dWithNorm(F1, D*F1, (nCh, 1), groups=F1, bias=bias, maxNorm=1),
            nn.BatchNorm2d(D*F1),
            nn.ELU(),
            nn.AvgPool2d((1, 8)),
            nn.Dropout2d(p)
        )
        
        self.separableConv = nn.Sequential(
            nn.Conv2d(D*F1, D*F1, (1, 16), padding='same', groups=D*F1, bias=bias),
            nn.Conv2d(D*F1, F2, 1, bias=bias),
            nn.BatchNorm2d(F2),
            nn.ELU(),
            nn.AvgPool2d((1, 16)),
            nn.Dropout2d(p)
        )
        
        self.flatten = nn.Flatten()
        self.fc = nn.Sequential(
            LinearWithNorm(self.get_size(nCh, nTime), cls, bias=bias, maxNorm=0.25),
            nn.LogSoftmax(dim=1)
        )
        
    def get_size(self, nCh, nTime):
        
        x = torch.randn(1, 1, nCh, nTime)
        out = self.filter(x)
        out = self.depthwiseConv(out)
        out = self.separableConv(out)
        out = self.flatten(out)
        
        return out.size(1)

    def forward(self, x):
        
        out = self.filter(x)
        out = self.depthwiseConv(out)
        out = self.separableConv(out)
        out = self.flatten(out)
        
        return self.fc(out)
    

if __name__ == '__main__':
    net = EEGNet(22, 1125).cuda()
    summary(net, (1, 1, 22, 1125))


#!/usr/bin/env python
# coding: utf-8

"""
    Util functions for dpeeg.
    
    @Author  : SheepTAO
    @Time    : 2023-09-15
"""


import os
import mne
import torch
import random
import functools
import numpy as np
from .tools.logger import Logger
from typing import (
    Any,
    Union,
    Callable,
    TypeVar,
)


# Life, the Universe, and Everything
DPEEG_SEED : int = 42
DPEEG_LOGGING_LEVEL = 'INFO'
DPEEG_DIR = os.path.join(os.path.expanduser('~'), 'dpeeg')


loger = Logger('dpeeg', flevel=None)


# Provide help for static type checkers:
# https://mypy.readthedocs.io/en/stable/generics.html#declaring-decorators
_FuncT = TypeVar('_FuncT', bound=Callable[..., Any])


def verbose(func : _FuncT) -> _FuncT:
    '''Verbose decorator to allow functions to override log-level.

    Parameters
    ----------
    func : callable
        Function to be decorated by setting the verbosity level.

    Returns
    -------
    dec : callable
        The decorated function.
    '''
    @functools.wraps(func)
    def inner(*args, **kwargs) -> _FuncT:
        kwargs.setdefault('verbose', DPEEG_LOGGING_LEVEL)
        loger._update_sh_level(kwargs['verbose'])
        # # for debug
        # print(kwargs, func)
        return func(*args, **kwargs)
    return inner


def set_log_level(
    verbose : Union[int, str] = DPEEG_LOGGING_LEVEL,
    retOldLevel : bool = False
) -> Union[None, str]:
    '''Set the global logging level.
    
    Parameters
    ----------
    verbose : int, str, optional
        The verbosity of messages to print. If a str, it can be either DEBUG,
        INFO, WARNING, ERROR, or CRITICAL. Note that these are for convenience
        and are equivalent to passing in logging.DEBUG, etc. If None, the 
        environment variable `DPEEG_LOGGING_LEVEL` is read, defaults to INFO.
    retOldLevel : bool, optional
        If True, return the old verbosity level. Default is False.

    Returns
    -------
    The old level. Only returned if `retOldLevel` is True.
    '''
    global DPEEG_LOGGING_LEVEL
    oldLevel = loger._get_sh_level()
    DPEEG_LOGGING_LEVEL = verbose
    mne.set_log_level(verbose)
    if retOldLevel:
        return oldLevel


def _set_random_seed(seed : int) -> None:
    '''Set the seed for Python's built-in random module and numpy.

    Parameters
    ----------
    seed: int
        The random seed to use.
    '''
    random.seed(seed)
    np.random.seed(seed)


def _set_torch_seed(seed : int) -> None:
    '''Set the seed for Python's built-in random module and numpy.

    Parameters
    ----------
    seed: int
        The random seed to use.

    '''
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)


def set_seed(
    seed : int = DPEEG_SEED, 
    retOldSeed : bool = False
) -> Union[None, int]:
    '''Set the seed for random, numpy and PyTorch.

    Parameters
    ----------
    seed : int, optional
        The random seed to use. If None, the environment variable `DPEEG_SEED` 
        is read, defaults to 42.
    retOldSeed : bool, optional
        If True, return the old seed. Default is False.

    Returns
    -------
    The old seed. Only returned if `retOldSeed` is True.
    '''
    global DPEEG_SEED
    oldSeed = DPEEG_SEED
    DPEEG_SEED = seed
    _set_random_seed(seed)
    _set_torch_seed(seed)
    if retOldSeed:
        return oldSeed
    


#!/usr/bin/env python
# coding: utf-8

"""
    @Desc    : A file handler to write in realtime.
    @Author  : SheepTAO
    @Time    : 2023-07-25
"""

import os


class Filer:
    def __init__(self, path : str, mode : str = 'a', exist_ok : bool = False) -> None:
        self.path = path
        self.mode = mode
        if os.path.exists(path) and not exist_ok:
            raise FileExistsError(f"File '{path}' already exists.")
        folder_path = os.path.dirname(path)
        os.makedirs(folder_path, exist_ok=True)

    def write(self, data : str) -> None:
        with open(self.path, self.mode) as f:
            f.write(data)


import csv


class CSVer(Filer):
    def __init__(self, path: str, mode: str = 'a') -> None:
        super().__init__(path, mode)

    def write(self, data: str) -> None:
        with open(self.path, self.mode) as f:
            w = csv.writer(f)
            w.writerows(data)



#!/usr/bin/env python
# coding: utf-8

"""
    This module is used to load commonly used MI-EEG data.
    
    NOTE: All data is read during the initialization phase.
        - Defect: At the beginning, all the data needs to be read once,
        and the initialization time is longer. If the amount of data is 
        huge, the memory consumption is huge.
        - Advantage: In the process of multiple iterations, there is no
        need to read data repeatedly, and only one round of IO operation
        is required, which saves time for multiple rounds of training.
        
    NOTE: If you want to build your own data set, you can refer to the
        sample dataset to build it. Please reassign the `_eventId` and
        `_raw` attributes. In the final stage of initialization, please
        call the `self._load_data()` to completely initialize your dataset.
        
    
    @Author  : SheepTAO
    @Time    : 2023-07-26
"""

import mne
from typing import Optional, List, Union

from .transforms import Transforms, ComposeTransforms, SplitTrainTest
from ..utils import loger, verbose, DPEEG_SEED


class EEGDataset:
    '''Base EEG dataset'''
    def _check_attr(self, attr : str) -> dict:
        '''Check if the attribute is None.'''
        if getattr(self, attr) is None:
            raise NotImplementedError(
                f"{self.__class__.__name__} model should implement '{attr}'.")
        return getattr(self, attr)
    
    @property
    def size(self) -> tuple:
        '''Return the current dimension of the dataset.'''
        sub = next(iter(self.dataset.values()))
        return tuple(sub['train'][0].shape)
        
    @property
    def eventId(self) -> dict:
        '''Return the event label of the dataset and its corresponding id.'''
        return self._check_attr('_eventId')
    
    @property
    def classes(self) -> tuple:
        '''Return the event label of the dataset.'''
        return tuple(self.eventId.keys())
    
    @property
    def raw(self) -> dict:
        '''Return the original Epochs data.'''
        return self._check_attr('_raw')
    
    @property
    def dataset(self) -> dict:
        '''Return the loaded data.'''
        return self._check_attr('_dataset')

    def __init__(
        self,
        transforms : Optional[Union[List[Transforms], Transforms]] = None,
        testSize : float = .25, 
        seed : int = DPEEG_SEED,
        verbose : Optional[str] = None, 
        verboseTrans : bool = False,
    ) -> None:
        '''EEG Dataset abstract base class.

        Parameters
        ----------
        subjects : list of int, optional
            List of subject number. If None, all subjects will be loaded. 
            Default is None.
        tmin, tmax : float
            Start and end time of the epochs in seconds, relative to the time-
            locked event. The closest or matching samples corresponding to the
            start and end time are included. Default is start and end time of 
            epochs, respectively.
        picks : list of str, optional
            Channels to include. If None, pick all channels. Default is None.
        transforms : list of Callable, optional
            List of pre-transforms on dataset. Default is None.
        testSize : float
            Split the training set and test set proportions. If the dataset is
            already split, it will be ignored. Default is 0.25.
        seed : int
            Random seed when splitting. Default is DPEEG_SEED.
        verbose : str, optional
            Log level of mne. Default is None.
        verboseTrans : bool
            If True, verbose of `transforms` will be set at the same time.
            Default is False.
        '''
        mne.set_log_level(verbose)
        
        self._testSize = testSize
        self._seed = seed
        self._transforms = transforms
        self._dataset = None
        self._verbose = verbose
        self._verboseTrans = verboseTrans
        
        # NOTE
        # Please make sure the following attributes are correctly overridden
        self._eventId = None        # task name and its corresponding label
        self._raw = None            # each subject and its corresponding Epochs
    
    def _load_data(self, split : bool = False) -> None:
        '''Extract data from Epochs and split.
        
        split : bool, optional
            Whether `self.raw` has been splited. Default is False.
        NOTE: Avoid data leakage when you split data.
        '''
        loger.info('Loading data from Epochs ...')
        dataset = {}
        for sub, sEpochs in self.raw.items():
            if not split:
                data = sEpochs.get_data()[:, :, :-1]
                label = sEpochs.events[:, -1]
                dataset[sub] = [data, label]
            else:
                dataset[sub] = {}
                for mode, mEpochs in sEpochs.items():
                    data = mEpochs.get_data()[:, :, :-1]
                    label = mEpochs.events[:, -1]
                    dataset[sub][mode] = [data, label]

        # split the dataset before transforms
        if self._transforms:
            trans = ComposeTransforms(self._transforms, self._verbose) \
                if self._verboseTrans else ComposeTransforms(self._transforms)
            if not split:
                trans.insert(0, SplitTrainTest(self._testSize, self._seed))
            self._dataset = trans(dataset)
        else:
            self._dataset = dataset
        loger.info('Loading dataset done.')
        
    
    def __getitem__(self, args) -> dict:
        '''Return the index position data of the corresponding subject
        and the corresponding set.
        
        Parameters
        ----------
        args : sub, index
        
        sub : int
            Number of subject.
        index : int, slice, optional
            The data index corresponding to the training set. If index is None,
            all the data (train and test) of the subject will be returned as a
            dictionary. Default is None.
        
        Returns
        -------
        dict - A dict contain train set (and test set).

        Example
        -------
            >>> dataset[1]
            {'train': (sub1[data], sub1[label]), 
             'test': (sub1[data], sub1[label])}
            >>> dataset[2, :2]
            (sub2[data[:2]], sub2[label[:2]])
        '''
        if isinstance(args, int):
            sub = args
            index = None
        else:
            num = len(args)
            if num != 2:
                raise ValueError(f'Index must be 1 to 2 numbers, but got {num}.')
            else:
                sub, index = args[0], args[1]
        
        if index:
            return {
                'train' : (self.dataset[sub]['train'][0][index],
                           self.dataset[sub]['train'][1][index])
            }
        else:
            return {
                'train': tuple(self.dataset[sub]['train']),
                'test': tuple(self.dataset[sub]['test'])
            }
        
    def __len__(self) -> int:
        return len(self.raw)
    
    def items(self):
        return self.dataset.items()
    
    def keys(self):
        return self.dataset.keys()
    
    def values(self):
        return self.dataset.values()
        
        
class PhysioNet(EEGDataset):
    '''Physionet MI Dataset.
    channels=64; subjects=1-109[88, 92, 100, 104], tasks={0:left hand, 1:right 
    hand, 2:hands, 3:feet}; duration=3s; freq=160Hz, sessions=1.
    '''
    @verbose
    def __init__(
        self,
        subjects : Optional[List[int]] = None,
        tmin : float = 0,
        tmax : float = 1,
        transforms : Optional[Union[List[Transforms], Transforms]] = None,
        testSize : float = .2,
        picks : Optional[List[str]] = None,
        baseline = None,
        seed : int = DPEEG_SEED,
        verbose : Optional[str] = None,
        verboseTrans: bool = False,
        **epoArgs
    ) -> None:
        '''Physionet MI Dataset.
        '''
        super().__init__(transforms, testSize, seed, verbose, verboseTrans)
        loger.info('Reading PhysionetMI Dataset ...')
        
        from moabb.datasets import PhysionetMI
        self._baseRaw = PhysionetMI().get_data(subjects)
        self._badSub = [88, 92, 100, 104]
        self._eventId = {
            'left_hand': 0,
            'right_hand': 1,
            'hands': 2,
            'feet': 3
        }
        
        self._raw = {}
        for sub, sess in self._baseRaw.copy().items():
            loger.info(f'\nCreate sub{sub} Epochs ...')
            if sub in self._badSub:
                continue
            
            epochsSes = []
            for run in sess['session_0'].values():
                
                ann = mne.events_from_annotations(run, regexp='^(?!rest$).*$')
                # update events (labels)
                ann[0][:, -1] = ann[0][:, -1]-1 if 'left_hand' in ann[1].keys() \
                    else ann[0][:, -1]+1
                # update event_id
                for key in ann[1].keys():
                    ann[1][key] = self._eventId[key]
                
                events = ann[0][:-1]
                epochsSes.append(
                    mne.Epochs(run, events, ann[1], tmin, tmax, baseline, picks,
                               preload=True, **epoArgs)
                )
            
            self._raw[sub] = mne.concatenate_epochs(epochsSes)
        
        self._load_data()
            

class BCICIV2A(EEGDataset):
    '''BCI Competition IV Data sets 2a.
    channels=22; subjects=1-9; tasks={0:left hand, 1:right hand, 2:feet, 3:tongue};
    duration=4s; freq=250Hz; sessions=2.
    '''
    @verbose
    def __init__(
        self,
        subjects : Optional[List[int]] = None,
        tmin : float = 0,
        tmax : float = 4,
        transforms : Optional[Union[List[Transforms], Transforms]] = None,
        testSize : float = .2,
        mode : int = 1,
        picks : Optional[List[str]] = None,
        baseline = None,
        seed : int = DPEEG_SEED,
        verbose : Optional[str] = None,
        verboseTrans: bool = False,
        **epoArgs
    ) -> None:
        '''BCI Competition IV Data sets 2a.

        Parameters
        ----------
        mode: int, optional
            If mode = 0, training data and test data will only use session 1.
            If mode = 1, training data and test data will use session 1 and 2, respectively.
            If mode = 2, training data and test data will use both session 1 and 2.
            Default is 1.
        '''
        super().__init__(transforms, testSize, seed, verbose, verboseTrans)
        loger.info('Reading BCICIV 2A Dataset ...')

        from moabb.datasets import BNCI2014001
        self._baseRaw = BNCI2014001().get_data(subjects)
        self._eventId = {
            'left_hand': 0,
            'right_hand': 1,
            'feet': 2,
            'tongue': 3
        }
        
        self._raw = {}
        for sub, data1 in self._baseRaw.copy().items():
            loger.info(f'Create sub{sub} Epochs ...')
            
            epochsSesOne, epochsSesTwo = [], []
            for session, data2 in data1.items():
                
                for run in data2.values():
                    
                    events = mne.find_events(run, 'stim')
                    events[:, -1] -= 1
                    epochs = mne.Epochs(run, events, self._eventId, tmin, tmax,
                                        baseline, picks, preload=True, **epoArgs)
                    epochs.drop_channels(['stim', 'EOG1', 'EOG2', 'EOG3'])
                    
                    if session == 'session_T':
                        epochsSesOne.append(epochs)
                    else:
                        epochsSesTwo.append(epochs)
            
            if mode == 0:
                self._raw[sub] = mne.concatenate_epochs(epochsSesOne)
            elif mode == 1:
                self._raw[sub] = {}
                self._raw[sub]['train'] = mne.concatenate_epochs(epochsSesOne)
                self._raw[sub]['test'] = mne.concatenate_epochs(epochsSesTwo)
            elif mode == 2:
                self._raw[sub] = mne.concatenate_epochs(epochsSesOne + epochsSesTwo)
            else:
                raise ValueError(f'Mode can only be 0, 1 and 2, but got {mode}.')
        
        split = True if mode == 1 else False
        self._load_data(split)


class HGD(EEGDataset):
    '''High Gamma Dataset.
    channels=128; subjects=1-14; tasks={0:feet, 1:left hand, 2:rest, 3:right hand};
    duration=4s; freq=500Hz; sessions=1.
    '''
    @verbose
    def __init__(
      self,
      subjects : Optional[List[int]] = None,
      tmin : float = 0,
      tmax : float = 4,
      transforms : Optional[Union[List[Transforms], Transforms]] = None,
      testSize : float = .2,
      picks : Optional[List[str]] = None,
      baseline = None,
      seed : int = DPEEG_SEED,
      verbose : Optional[str] = None,
      verboseTrans: bool = False,
      **epoArgs,
    ) -> None:
        '''High Gamma Dataset.
        '''
        super().__init__(transforms, testSize, seed, verbose, verboseTrans)
        loger.info('Reading High Gamma Dataset ...')

        from moabb.datasets import Schirrmeister2017
        self._baseRaw = Schirrmeister2017().get_data(subjects)
        self._eventId = {
            'feet': 0, 
            'left_hand': 1, 
            'rest': 2, 
            'right_hand': 3
        }

        self._raw = {}
        for sub, sess in self._baseRaw.items():
            loger.info(f'Create sub{sub} Epochs ...')
            for mode, run in sess['session_0'].items():
                events, _ = mne.events_from_annotations(run)
                # update events
                events[:, -1] -= 1
                epochs = mne.Epochs(run, events, self._eventId, tmin, tmax,
                                    baseline, picks, preload=True, **epoArgs)
                self._raw.setdefault(sub, {})[mode] = epochs

        self._load_data(split=True)



from .datasets import (
    EEGDataset,
    PhysioNet,
    BCICIV2A,
    HGD,
)
from .functions import (
    split_train_test,
    to_tensor,
    slide_win,
    save,
    load,
)
from .transforms import (
    Transforms,
    ComposeTransforms,
    SplitTrainTest,
    ToTensor,
    Normalization,
    SlideWin,
    ApplyFunc,
    Save,
    
)


#!/usr/bin/env python
# coding: utf-8

"""
    This module defines with what experiments the module is trained and validated.

    experiments:
        1 - kFoldCV : K-Fold cross-validation.
        2 - HoldOut : hold out validation.

    @Author  : SheepTAO
    @Time    : 2023-07-26
"""


import os, torch
from torch import Tensor
from typing import Union, Optional, Tuple
from sklearn.model_selection import StratifiedKFold
from torchmetrics.functional.classification.cohen_kappa import cohen_kappa

from .train import Train
from ..data.datasets import EEGDataset
from ..tools import Timer, Filer, Logger
from ..utils import DPEEG_SEED, DPEEG_DIR
from .evaluate import save_cm_img


class KFoldCV:
    def __init__(
        self,
        trainer : Train,
        k : int = 10,
        outFolder : Optional[str] = None,
        maxEpochs_1 : int = 1500,
        maxEpochs_2 : int = 600,
        noIncreaseEpochs : int = 100,
        varCheck : str = 'valInacc',
        shuffle : bool = True,
        seed : int = DPEEG_SEED,
        verbose : Union[int, str] = 'INFO',
    ) -> None:
        '''K-Fold cross-validation class.

        Parameters
        ----------
        trainer : Train
            Trainer used for training module on dataset.
        k : int, optional
            k of k-Fold. Default is 10.
        outFolder : str, optional
            Store all results during experiments to the given folder. Default is
            '~/dpeeg/out/model_name/'.
        maxEpochs_1, maxEpochs_2 : int
            Maximum number of epochs in the x stage of training. Default is
            1500 and 600 respectively.
        noIncreaseEpochs : int
            Maximum number of consecutive epochs when the accuracy of the first-
            stage validation set has no relative improvement. Default is 100.
        varCheck : str
            The best value ('valInacc'/'valLoss') to check while determining 
            the best model which will be used for parameter initialization in
            the second stage of model training. Default is 'valInacc'.
        shuffle : bool
            Wheter shuffle dataset. Default is True.
        seed : int
            Seed of random for review. Default is DPEEG_SEED.
        verbose : int, str
            The log level of console. Default is INFO. Mainly used for debugging.

        All experimental results of the model will be saved in the outFolder
        directory of the trainer parameter.
        TODO: Nested Cross-Validation.
        '''
        self.trainer = trainer
        self.k = k
        self.maxEpochs_1 = maxEpochs_1
        self.maxEpochs_2 = maxEpochs_2
        self.noIncreaseEpochs = noIncreaseEpochs
        self.varCheck = varCheck
        
        self.loger = Logger('dpeeg_exp', clevel=verbose)
        self.timer = Timer()
        
        # set output folder
        self.outFolder = os.path.abspath(outFolder) if outFolder else \
            os.path.join(DPEEG_DIR, 'out', trainer.net.__class__.__name__)
        os.makedirs(self.outFolder, exist_ok=True)
        self.loger.info(f'Results will be saved in folder: {self.outFolder}')

        # create a stratified k-fold index
        self.skf = StratifiedKFold(k, shuffle=shuffle, random_state=seed)

    def kfold(
        self,
        trainset : Union[tuple, list],
        testset : Union[tuple, list],
        clsName : Union[tuple, list],
        sub : Optional[str] = None,
    ) -> Tuple[Tensor, ...]:
        '''Basic K-Fold cross-validation function.

        Parameters
        ----------
        trainset: tuple, list
            Dataset used for training module. If type is tuple or list, dataset
            should be (data, labels).
        testset:
            Dataset used for evaluating moduel. If type is tuple or list, dataset
            should be (data, labels).
        sub : str, optional
            Create a subdirectory of outFolder to store all yield results during 
            KFold_CV. If None, will use outFolder. Default is None.
        clsName : tuple, list
            The name of dataset labels.
        
        Returns
        -------
        Return trainAcc, valAcc, testAcc and testKappa.
        '''
        def get_res(
            results : dict, 
            task : str, 
            var : str = 'acc'
        ) -> Tensor:
            '''Get all value from the results by var and return them.
            '''
            res = []
            for i in range(self.k):
                data = results[task][i][var].unsqueeze(0) if var == 'acc' \
                    else results[task][i][var]
                res.append(data)
            return torch.cat(res)

        # save trained results
        results = {
            'train' : [],
            'val'   : [],
            'test'  : [],
        }

        # set kFold-cv results storage path
        subFolder = os.path.join(self.outFolder, sub) if sub else self.outFolder
        os.makedirs(subFolder, exist_ok=True)

        # store experiment results
        resPath = os.path.join(subFolder, 'resutls.txt')
        filer = Filer(resPath)
        cmFolder = os.path.join(subFolder, 'confusion_matrix')
        os.makedirs(cmFolder)
        
        # register a timer
        self.timer.start('kfold')

        # run k-fold
        for idx, (trainIdx, valIdx) in enumerate(self.skf.split(*trainset)):
            self.loger.info(f'\n# ------ {sub} - expNo.{idx+1} ------ #')

            # train one fold
            trainData = (trainset[0][trainIdx], trainset[1][trainIdx])
            valData = (trainset[0][valIdx], trainset[1][valIdx])
            expPath = os.path.join(subFolder, f'expNo_{idx + 1}')
            res = self.trainer.fit_with_val(
                trainData, valData, testset, expPath, clsName, self.maxEpochs_1,
                self.maxEpochs_2, self.noIncreaseEpochs, self.varCheck
            )
            trainRes, valRes, testRes = res['train'], res['val'], res['test']

            # save the confusion matirx image
            save_cm_img(testRes['preds'], testRes['target'], clsName,
                        os.path.join(cmFolder, f'expNo_{idx+1}_CM.png')
            )

            # save all results
            results['train'].append(trainRes)
            results['val'].append(valRes)
            results['test'].append(testRes)

            filer.write(f'------------------ ExpNo_{idx+1} ------------------\n')
            filer.write(f'Acc : Train={trainRes["acc"]:.4f} | Val={valRes["acc"]:.4f}' +
                        f' | Test={testRes["acc"]:.4f}\n\n')
        
        h, m, s = self.timer.stop('kfold')
        self.loger.info(f'\n[Cross-Validation Finish]')
        self.loger.info(f'Cost time = {h}H:{m}M:{s:.2f}S')

        # calculate the acerage acc
        trainAccList = get_res(results, 'train')
        valAccList = get_res(results, 'val')
        testAccList = get_res(results, 'test')

        trainAcc = torch.mean(trainAccList)
        valAcc = torch.mean(valAccList)
        testAcc = torch.mean(testAccList)
        self.loger.info(f'Acc : train={trainAcc:.4f} | val={valAcc:.4f} | test={testAcc:.4f}') 
        
        # calculate cohen kappa
        testKappa = cohen_kappa(
            get_res(results, 'test', 'preds'),
            get_res(results, 'test', 'target'),
            task = 'multiclass',
            num_classes = len(clsName)
        )

        # store the subject accuracy and cohen kappa
        filer.write(f'------------- Average -------------\n')
        filer.write(f'Acc = {testAcc*100:.2f}\u00b1{torch.std(testAccList)*100:.2f}% | '+
                    f'Kappa = {testKappa:.2f}\n')

        return trainAcc, valAcc, testAcc, testKappa

    def run(
        self, 
        dataset : Union[EEGDataset, dict],
        clsName : Union[tuple, list], 
        datasetName : Optional[str] = None,
        desc : Optional[str] = None,
    ) -> None:
        '''Run K-Fold cross validation on eeg datasets.

        Will select train set to process k-fold cross validation `kFold`, and 
        test set to evaluate the model.

        Parameters
        ----------
        dataset : EEGDataset, dict
            Dataset used for k-fold. Dataset structured as:
            
            {0 : { 'train': (array(nsample x channels x times), array(nsample)),
                   'test': (array(nsample x channels x times), array(nsample))},
             1 : {...},
            ...}
            
            where array can be `numpy.ndarray` or `torch.Tensor`. Please make sure 
            that the dataset has been split.
        clsName : tuple, list
            The name of dataset labels.
        datasetName : str, optional
            The dataset name to use. If None, `dataset.__class__.__name__` will be
            used as the folder to save experimental results.
        desc : str, optional
            Add a short description to the current experiment. Default is None.
        '''
        self.loger.info(f'\n[{self.k}-Fold Cross-Validation Starting]')
        
        # store all subjects' results in one file
        self.outFolder = os.path.join(self.outFolder, dataset.__class__.__name__) \
            if not datasetName else os.path.join(self.outFolder, datasetName)
        os.makedirs(self.outFolder)
        resPath = os.path.join(self.outFolder, 'resutls.txt')
        filer = Filer(resPath)
        filer.write(f'[Start Time]: {self.timer.ctime()}\n\n')
        filer.write(f'[ExP Description]: {desc}\n\n')
        filer.write(str(self.trainer) + '\n\n')
            
        # resport results
        testAccList = []
        testKappaList = []

        # update root timer and start k-fold cross validation for each subject
        self.timer.start()
        self.loger.info('=' * 50)
        for sub, data in dataset.items():

            self.loger.info(f'\n[Subject-{sub} Training ...]')
            self.loger.info(f'Train set = {data["train"][1].shape} | Test set = '
                  f'{data["test"][1].shape}')
            self.loger.info('-' * 50)

            # run a k-Fold cross validation
            _, _, testAcc, testKappa = \
                self.kfold(data['train'], data['test'], clsName, f'sub{sub}')
            
            testAccList.append(testAcc.unsqueeze(0))
            testKappaList.append(testKappa.unsqueeze(0))

            filer.write(f'---------- Sub_{sub} ----------\n')
            filer.write(f'Acc = {testAcc*100:.2f}% | Kappa = {testKappa:.2f}\n\n')

        acc = torch.mean(torch.cat(testAccList))
        std = torch.std(torch.cat(testAccList))
        kappa = torch.mean(torch.cat(testKappaList))
        # store the model accuracy and cohen kappa
        filer.write(f'------------- MODEL ------------\n')
        filer.write(f'Acc = {acc*100:.2f}\u00b1{std*100:.2f}% | Kappa = '
                    f'{kappa:.2f}\n')

        # end
        h, m, s = self.timer.stop()
        self.loger.info(f'\n[All subjects finished]')
        self.loger.info(f'Cost time = {h}H:{m}M:{s:.2f}S')
        self.loger.info('=' * 50)
        self.loger.info(f'[Acc = {acc*100:.2f}\u00b1{std*100:.2f}% | Kappa = '
                        f'{kappa:.2f}]')
        

class HoldOut:
    '''TODO'''
    def __init__(self) -> None:
        pass



#!/usr/bin/env python
# coding: utf-8

"""
    A stopping criteria class to compose any training termination condition.

    @Author  : SheepTAO
    @Time    : 2023-07-26
"""

import abc
import os, sys, json
from typing import Union


CURRENT_MODULE = sys.modules[__name__]


class Criteria(abc.ABC):
    @abc.abstractmethod
    def __call__(self, variables : dict) -> bool:
        pass

    @abc.abstractmethod
    def __repr__(self) -> str:
        pass

    def reset_parameters(self) -> None:
        pass


class ComposeStopCriteria(Criteria):
    '''Advanced complex logical expression parser to compose stopping criteria.
    '''
    def __init__(self, stopcri : Union[dict, str]) -> None:
        '''The parent class for all the stop criteria.
        
        Parameters
        ----------
        stopcri : dict, str
            Type is str, indicate the configuration file path. Type is dict, 
            indicate a dictionary containing initialization parameters for the
            corresponding stopping criteria.
        
        Returns
        -------
            Whether the stopping condition is reached.
        '''
        if isinstance(stopcri, str):
            with open(os.path.abspath(stopcri)) as fp:
                cri : dict = json.load(fp)
        else:
            cri : dict = stopcri
        self.call : Criteria = CURRENT_MODULE.__dict__[list(cri.keys())[0]] \
            (**cri[list(cri.keys())[0]])

    def __call__(self, variables: dict) -> bool:
        return self.call(variables)

    def __repr__(self) -> str:
        return str(self.call)

    def reset_parameters(self) -> None:
        self.call.reset_parameters()


class And(Criteria):
    def __init__(self, cri1, cri2) -> None:
        '''And operation on two stop criteria.

        Parameters
        ----------
        cri1 : dict
            Dictionary describing first criteria.
        cri2 : dict
            Dictionary describing second criteria.

        Notes
        -----
        If you wish to do and on multiple cases then do like: And(And(A, B), C)...
        '''
        self.cri1 : Criteria = CURRENT_MODULE.__dict__[list(cri1.keys())[0]] \
            (**cri1[list(cri1.keys())[0]])
        self.cri2 : Criteria = CURRENT_MODULE.__dict__[list(cri2.keys())[0]] \
            (**cri2[list(cri2.keys())[0]])

    def __call__(self, variables: dict) -> bool:
        return self.cri1(variables) and self.cri2(variables)

    def __repr__(self) -> str:
        return f'({self.cri1} \'AND\' {self.cri2})'

    def reset_parameters(self) -> None:
        self.cri1.reset_parameters()
        self.cri2.reset_parameters()


class Or(Criteria):
    def __init__(self, cri1 : dict, cri2 : dict) -> None:
        '''Or operation on two stop criteria.

        Parameters
        ----------
        cri1 : dict
            Dictionary describing first criteria.
        cri2 : dict
            Dictionary describing second criteria.

        Notes
        -----
        If you wish to do and on multiple cases then do like: And(And(A, B), C)...
        '''
        self.cri1 : Criteria = CURRENT_MODULE.__dict__[list(cri1.keys())[0]] \
            (**cri1[list(cri1.keys())[0]])
        self.cri2 : Criteria = CURRENT_MODULE.__dict__[list(cri2.keys())[0]] \
            (**cri2[list(cri2.keys())[0]])

    def __call__(self, variables: dict) -> bool:
        return self.cri1(variables) or self.cri2(variables)

    def __repr__(self) -> str:
        return f'({self.cri1} \'OR\' {self.cri2})'

    def reset_parameters(self) -> None:
        self.cri1.reset_parameters()
        self.cri2.reset_parameters()


class MaxEpoch(Criteria):
    def __init__(self, maxEpochs : int, varName : str) -> None:
        '''Stop when given number of epochs reached.

        Parameters
        ----------
        maxEpochs : int
            Maximum epochs to watch. 
        varName : str
            Key name to compare with in the variables dictionary.
        '''
        self.maxEpochs = maxEpochs
        self.varName = varName

    def __call__(self, variables : dict) -> bool:
        return variables[self.varName] >= self.maxEpochs

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}(maxEpochs={self.maxEpochs}'
                f', varName={self.varName})')


class NoDecrease(Criteria):
    def __init__(self, numEpochs : int, varName : str, minChange = 1e-6) -> None:
        '''Stop if there is no decrease on a given monitor channel for given 
        number of epochs.

        Parameters
        ----------
        numEpochs : int
            Number of epochs to wait while there is no decrease in the value.
        varName : str
            Key name to compare with in the variables dictionary.
        minChange : float
            Minimum relative decrease which resets the numEpochs. Default is 1e-6.
        '''
        self.numEpochs = numEpochs
        self.varName = varName
        self.minChange = minChange
        self.minValue = float('inf')
        self.currentEpoch = 0

    def __call__(self, variables : dict) -> bool:
        if variables[self.varName] <= (1 - self.minChange) * self.minValue:
            self.minValue= variables[self.varName]
            self.currentEpoch = 1
        else:
            self.currentEpoch += 1

        return self.currentEpoch >= self.numEpochs

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}(numEpochs={self.numEpochs}'
                f', varName={self.varName}, minChange={self.minChange})')

    def reset_parameters(self) -> None:
        self.minValue = float('inf')
        self.currentEpoch = 0


class Bigger(Criteria):
    def __init__(self, var : float, varName : str) -> None:
        '''Stop when greater than the specified value.

        Parameters
        ----------
        var : float
            Maximum value to watch.
        varName : str
            Key name to compare with in the variables dictionary.
        '''
        self.var = var
        self.varName = varName

    def __call__(self, variables: dict) -> bool:
        return variables[self.varName] > self.var

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(var={self.var})'


class Smaller(Criteria):
    def __init__(self, var : float, varName : str) -> None:
        '''Stop when less than the specified value.

        Parameters
        ----------
        var : float
            Minimum value to watch.
        varName : str
            Key name to compare with in the variables dictionary.
        '''
        self.var = var
        self.varName = varName

    def __call__(self, variables: dict) -> bool:
        return variables[self.varName] < self.var

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(var={self.var})'



#!/usr/bin/env python
# coding: utf-8

"""
    This module is used to perform common preprocessing on eeg data.

    TODO: Data augmentation.
    
    @author: SheepTAO
    @data: 2023-5-15
"""


import abc
import numpy as np
import pandas as pd
from typing import Optional, Callable, Union, List

from ..utils import loger, verbose, DPEEG_SEED
from .functions import (
    split_train_test,
    to_tensor,
    slide_win,
    save,
)


class Transforms(abc.ABC):
    def __init__(self, verbose : Optional[Union[int, str]] = None) -> None:
        self.verbose = verbose
    
    @abc.abstractmethod
    def __call__(self, input : dict) -> dict:
        pass

    @abc.abstractmethod
    def __repr__(self) -> str:
        pass


class ComposeTransforms(Transforms):
    '''Composes several transforms together.
    '''
    @verbose
    def __init__(
        self, 
        transforms : Union[List[Transforms], Transforms], 
        verbose : Optional[Union[int, str]] = None,
    ) -> None:
        '''Composes several transforms together. 
        The transforms in `ComposeTransforms` are connected in a cascading way.

        Parameters
        ----------
        transforms : list of Transforms, Transforms
            Transforms (list of `Transform` objects): list of transforms to compose.
            If is Transforms, then will be turned into a list containing input.
        verbose : int, str, optional
            The log level of the entire transformation list. Default is None (INFO).
        '''
        super().__init__(verbose)
        self.trans = [transforms] \
            if isinstance(transforms, Transforms) else transforms
        for tran in self.trans:
                tran.verbose = verbose
        
    def __call__(self, input):
        loger.info('[Transform dataset ...]')
        loger.info('---------------------')
        for t in self.trans:
            input = t(input)
        loger.info('---------------')
        loger.info('[Transform done.]')
        return input

    def __repr__(self) -> str:
        s = 'Compose('
        if len(self.trans) == 0:
            return s + ')'
        else:
            for idx, tran in enumerate(self.trans):
                s += f'\n ({idx}): {tran}'
        return s + '\n)'

    def appends(self, transforms):
        '''Append a transform or a list of transforms to the last of composes.
        '''
        if isinstance(transforms, list):
            self.trans.extend(transforms)
        else:
            self.trans.append(transforms)
        
    def insert(self, index, transform):
        '''Insert a transform at index.'''
        self.trans.insert(index, transform)


class SplitTrainTest(Transforms):
    '''Split the dataset into training and testing sets.
    '''
    @verbose
    def __init__(
        self, 
        testSize : float = .25, 
        seed : int = DPEEG_SEED, 
        sample : Optional[List[int]] = None,
        verbose : Optional[Union[int, str]] = None
    ) -> None:
        '''Split the dataset into training and testing sets.

        Parameters
        ----------
        testSize : float
            The proportion of the test set. Default is 0.25. If index is not None,
            testSize will be ignored. Default use stratified fashion and the last
            arr serves as the class labels.
        seed : int
            Random seed when splitting. Default is DPEEG_SEED.
        sample : list of int, optional
            A list of integers, the entries indicate which data were selected
            as the test set. If None, testSize will be used. Default is None.
        '''
        super().__init__(verbose)
        self.testSize = testSize
        self.seed = seed
        self.sample = sample
        self.verbose = verbose

    def __call__(self, input : dict) -> dict:
        loger.info(f'[{self} starting ...]')
        for sub, data in input.items():
            trainX, testX, trainy, testy = split_train_test(
                data[0], data[1], testSize=self.testSize, seed=self.seed,
                sample=self.sample, verbose=self.verbose
            )
            input[sub] = {}
            input[sub]['train'] = [trainX, trainy]
            input[sub]['test'] = [testX, testy]
        return input

    def __repr__(self) -> str:
        s = 'SplitDataset('
        if self.sample:
            s += 'sample'
        else:
            s += f'testSize={self.testSize}'
            if self.seed:
                s += f', seed={self.seed}'
        return s + ')'


class ToTensor(Transforms):
    '''Convert the numpy data in the dataset into Tensor format.
    '''
    @verbose
    def __init__(self, verbose) -> None:
        super().__init__(verbose)
        self.verbose = verbose

    def __call__(self, input : dict) -> dict:
        loger.info(f'[{self} starting ...]')
        for sub in input.values():
            for mode in ['train', 'test']:
                sub[mode][0], sub[mode][1] = to_tensor(
                    sub[mode][0], sub[mode][1], verbose=self.verbose
                )
        return input

    def __repr__(self) -> str:
        return 'ToTensor'


class Normalization(Transforms):
    '''Normalize the data.
    '''
    @verbose
    def __init__(
        self, 
        mode : str = 'z-score', 
        factorNew : float = 1e-3,
        eps : float = 1e-4,
        verbose : Optional[Union[int, str]] = None
    ) -> None:
        '''Normalize data in the given way in the given dimension.

        Parameters
        ----------
        mode : str
            within subject:
            - `z-score`,
                :math: $X_{i}=\\frac{X_{i}-mean(X_{i})}{std(X_{i})}$ where mean
                represents the channel-wise average value and std represents
                the channel-wise standard deviation value, calculated with the 
                training data and used directly for the test data.
            - `ems` (exponential moving standardize),
                Compute the exponental moving mean :math: $m_t$ at time $t$ as
                :math: $m_t=\\mathrm{factornew}\\cdot mean(x_t)+(1-\\mathrm{factornew})\\cdot m_{t-1}$.
                Then compute exponential moving variance $v_t$ at time t as
                :math: $v_t=\\mathrm{factornew}\\cdot (m_t-x_t)^2+(1-\\mathrm{factornew})\\cdot v_{t-1}$.
                Finally, standardize the data point :math: $x_t$ at time $t$ as:
                :math: $x'_t=(x_t-m_t)/max(\\sqrt{->v_t}, eps)$.
            cross subject:
            - `ea` (euclidean-space alignment),
                An unsupervised standardization of data transfer across subjects.
                > H. He and D. Wu, 'Transfer Learning for Brain-Computer Interfaces: A Euclidean Space 
                Data Alignment Approach', Ieee T Bio-med Eng, vol. 67, no. 2, pp. 399-410, Feb. 2020, 
                doi: 10.1109/TBME.2019.2913914.

            Default is z-score.
            
        factorNew : float
            Smoothing factor of exponential moving standardize. Default is 1e-3.
        eps : float
            Stabilizer for division by zero variance. Default is 1e-4.
        '''
        super().__init__(verbose)
        self.modeList = ['z-score', 'ems', 'ea']
        self.mode = mode
        self.factorNew = factorNew
        self.eps = eps
        self.verbose = verbose

    def __call__(self, input : dict) -> dict:
        if self.mode not in self.modeList:
            raise ValueError('Only the following normalization methods are '+
                             f'supported: {self.modeList}')

        loger.info(f'[{self} starting ...]')
        R = np.empty(0)
        if self.mode == 'ea':
            dataList = []
            for sub in input.values():
                for mode in ['train', 'test']:
                    dataList.append(sub[mode][0] / np.max(np.abs(sub[mode][0])))
            data = np.concatenate(dataList)

            Rbar = np.zeros((data.shape[1], data.shape[1]))
            for i in range(len(data)):
                Rbar += np.dot(data[i], data[i].T)
            RbarMean = Rbar / len(data)
            R = np.linalg.inv(np.sqrt(RbarMean))

        for sub in input.values():

            if self.mode == 'z-score':
                m, s = np.mean(sub['train'][0]), np.std(sub['train'][0])
                for mode in ['train', 'test']:
                    sub[mode][0] = (sub[mode][0] - m) / np.maximum(self.eps, s)

            # BUG
            elif self.mode == 'ems':
                df = pd.DataFrame(sub['train'][0].T)
                meaned = df.ewm(alpha=self.factorNew).mean()
                for mode in ['train', 'test']:
                    df = pd.DataFrame(sub[mode][0].T)
                    demeaned = df - meaned
                    squared = demeaned * demeaned
                    squareEwmed = squared.ewm(alpha=self.factorNew).mean()
                    standardized = demeaned / np.maximum(self.eps, 
                                            np.sqrt(np.array(squareEwmed)))
                    standardized = np.array(standardized)
                    sub[mode][0] = standardized.T

            # NOT TEST
            elif self.mode == 'ea':
                for mode in ['train', 'test']:
                    sub[mode][0] /= np.max(np.abs(sub[mode][0]))
                    sub[mode][0] = np.dot(sub[mode][0], R)
        return input

    def __repr__(self) -> str:
        s = f'Normalization(mode={self.mode}'
        if self.mode == 'z-score':
            return s + ')'
        elif self.mode == 'ems':
            return s + f', factorNew={self.factorNew}, eps={self.eps})'
        return s + ')'


class SlideWin(Transforms):
    '''Apply a sliding window to the dataset.
    '''
    @verbose
    def __init__(
        self, 
        win : int = 125, 
        overlap : int = 0,
        verbose : Optional[Union[int, str]] = None
    ) -> None:
        '''This transform is only splits the time series (dim = -1) through the sliding 
        window operation on the original dataset. If the time axis is not divisible by 
        the sliding window, the last remaining time data will be discarded.

        Parameters
        ----------
        win : int
            The size of the sliding window.
        overlap : int
            The amount of overlap between adjacent sliding windows. Default is 0.
        '''
        super().__init__(verbose)
        self.win = win
        self.overlap = overlap
        self.verbose = verbose

    def __call__(self, input : dict) -> dict:
        loger.info(f'[{self} starting ...]')
        for sub, data in input.items():
            loger.info(f'Sliding window to sub_{sub}.')
            for mode in ['train', 'test']:
                data[mode][0], data[mode][1] = slide_win(
                    data[mode][0], self.win, self.overlap,
                    data[mode][1], verbose=self.verbose
                )
        return input

    def __repr__(self) -> str:
        s = f'SlideWin(win={self.win}'
        if self.overlap != 0:
            s += f', overlap={self.overlap}'
        return s + ')'


class ApplyFunc(Transforms):
    '''Apply a function on data.
    '''
    @verbose
    def __init__(
        self, 
        func : Callable, 
        verbose: Optional[Union[int, str]] = None
    ) -> None:
        '''This transform can be used to filter the data (via the `mne` library or
        other methods), change the data shape (via the `numpy`) and so on.

        NOTE:
        If you want to pass a function with parameters, such as you want to
        use `np.expand_dims()` with `axis` parameter, you can do as follows:
        >>> from functools import partial
        >>> transforms.ApplyFunc(partial(np.expand_dims, axis=1))
        or
        >>> transforms.ApplyFunc(lambda x: np.expand_dims(x, 1))
        '''
        super().__init__(verbose)
        self.func = func
        self.verbose = verbose

    def __call__(self, input : dict) -> dict:
        for sub in input.values():
            for mode in ['train', 'test']:
                sub[mode][0] = self.func(sub[mode][0])
        return input

    def __repr__(self) -> str:
        return f'ApplyFunc(func={self.func})'


class Save(Transforms):
    '''Save the transformed data.
    '''
    @verbose
    def __init__(
        self, 
        folder : str, 
        verbose : Optional[Union[int, str]] = None
    ) -> None:
        '''Save transformed dataset to a binary file in NumPy `.npy` format.

        Parameters
        ----------
        folder : str
            Folder name to save transformed data.
        '''
        super().__init__(verbose)
        self.folder = folder
        self.verbose = verbose

    def __call__(self, input : dict) -> None:
        loger.info(f'[{self} starting ...]')
        save(self.folder, input, verbose=self.verbose)

    def __repr__(self) -> str:
        return f'Save(folder={self.folder})'



#!/usr/bin/env python
# coding: utf-8

"""
    The train model for any deep learning analysis.
    This class should provide following functionalities for any deep learning.

    module:
        1. train() -> Train the model
        2. predict() -> Evaluate the train, validation, and test performance
        3. Create train and validation graphs
        4. Run over CPU / GPU (if available)
    This class needs following things to run:
        1. net -> The architecture of the network. It should inherit nn.Module
             and should define the forward method.
        2. trainData, testData and valData -> these should be tuple, list or
             DataLoder the first parameters is `data` and second is `label`.
        3. optimizer -> the optimizer of type torch.optim.
        4. lr_scheduler -> the scheduler of type torch.optim.lr_scheduler.
        5. outFolder -> the folder where the results will be stored.
        6. nGPU -> will run on GPU only if it's available
             use CPU when the GPU is not available. 

    NOTE: The train model only support single-card training.

    @Author  : SheepTAO
    @Time    : 2023-07-24
"""


import os, torch, pickle
from torch import nn
from torch import optim
from torch import Tensor
from copy import deepcopy
from torchinfo import summary
from typing import Optional, Tuple, Union, Dict
from torch.utils.tensorboard.writer import SummaryWriter
from torch.utils.data import TensorDataset, DataLoader, ConcatDataset
from torchmetrics.functional.classification.accuracy import accuracy

from ..tools import Logger, Timer
from ..utils import DPEEG_SEED
from ..data.functions import to_tensor
from .stopcriteria import ComposeStopCriteria

from torch.backends import cudnn
cudnn.benchmark = False
cudnn.deterministic = True


# base train Model
class Train:
    '''Apex Trainer for any deep learning.
    '''
    def __init__(
        self,
        net : nn.Module,
        nGPU : int = 0,
        seed : int = DPEEG_SEED,
        lossFn : str = 'NLLLoss',
        lossFnArgs : dict = {},
        optimizer : str = 'AdamW',
        optimArgs : dict = {},
        lr : float = 1e-3,
        lrSch : Optional[str] = None,
        lrSchArgs : dict = {},
        gradAcc : int = 1,
        batchSize : int = 256,
        dataSize : Optional[Union[tuple, list]] = None,
        depth : int = 3,
        verbose : Union[int, str] = 'INFO',
    ) -> None:
        '''Initialize the basic attribute of the train model.
        Generate a trainer to test the performance of the same network on different
        datasets.

        Parameters
        ----------
        net : nn.Module
            Inherit nn.Module and should define the forward method.
        nGPU : int
            Select the gpu id to train. Default is 0.
        seed : int
            Select random seed for review. Default is DPEEG_SEED.
        lossFn : str
            Name of the loss function from torch.nn which will be used for
            training. Default is NLLLoss.
        lossFnArgs : dict
            Additional arguments to be passed to the loss function. Default is {}.
        optimizer : str
            Name of the optimization function from torch.optim which will be used
            for training. Default is AdamW.
        optimArgs : dict
            Additional arguments to be passed to the optimization function.
            Default is {}.
        lr : float
            Learning rate. Default is 1e-3.
        lrSch : str
            Name of the lr_scheduler from torch.optim.lr_scheduler which will be 
            used for training. Default is ''.
        lrSchArgs : dict
            Additional arguments to be passed to the lr_scheduler function.
            Default is {}.
        gradAcc : int
            Aradient accumulation. Default is 1.
        batchSize : int
            Mini-batch size. Default is 256.
        dataSize : tuple, list, optional
            Output the structure of the network model according to the input data
            dimension if the `dataSize` is given. Default is None.
        depth : int
            Depth of nested layers to display. Default is 3.
        verbose : int, str
            The log level of console. Default is INFO. Mainly used for debugging.
        '''
        self.net = net

        self.loger = Logger('dpeeg_train', clevel=verbose)
        self.timer = Timer()

        # init trainer
        self.device = self.get_device(nGPU)
        self.net.to(self.device)
        self.set_seed(seed)

        # summarize network structure
        self.netArch = str(net) + '\n'
        if dataSize:
            self.netArch += str(summary(net, dataSize, depth=depth))
        self.loger.info(self.netArch)

        # the type of optimizer, etc. selected
        self.lossFnType = lossFn
        self.optimizerType = optimizer
        self.lrSchType = lrSch

        # save additional parameters
        self.lossFnArgs = lossFnArgs
        self.optimArgs = optimArgs
        self.lrSchArgs = lrSchArgs
        # --- others
        self.lr = lr
        self.seed = seed
        self.gradAcc = gradAcc
        self.batchSize = batchSize
        self.verbose = verbose

        # set experimental details
        self.expDetails = {'expParam': {'seed': seed, 'lossFn': lossFn, 
            'lossFnArgs': lossFnArgs, 'optimizer': optimizer, 'lr': lr,
            'optimArgs': optimArgs, 'lrSch': lrSch, 'lrSchArgs': lrSchArgs,
            'batchSize': batchSize, 'gradAcc': gradAcc},
            'origNetParam': deepcopy(self.net.state_dict()),
        }
    
    def fit_with_val(
        self,
        trainData : Union[tuple, list],
        valData : Union[tuple, list],
        testData : Union[tuple, list],
        logDir : str,
        clsName : Union[tuple, list],
        maxEpochs_1 : int = 1500,
        maxEpochs_2 : int = 600,
        noIncreaseEpochs : int = 100,
        varCheck : str = 'valInacc',
    ) -> Dict[str, Dict[str, Tensor]]:
        '''Two-stage training strategy was used. In the first stage, the model 
        was trained using only the training set with the early stopping criteria
        whereby the validation set accuracy and loss was monitored and training
        was stopped if there was no increase in the validation set accuracy (or 
        loss) for consecutive 200 epochs. After reaching the stopping criteria, 
        network parameters with the best validation set accuracy (or loss) were
        restored. In the second stage, the model was trained with the complete 
        training data (train + validation set). The second stage training was
        stopped when the validation set loss reduced below the stage 1 training 
        set loss.

        Parameters
        ----------
        trainData : tuple, list
            Dataset used for training. If type is tuple or list, dataset should
            be (data, labels).
        valData : tuple, list
            Dataset used for validation. If type is tuple or list, dataset 
            should be (data, labels).
        testData : tuple, list
            Dataset used to evaluate the model. If type is tuple or list, 
            dataset should be (data, labels).
        logDir : str
            Directory location and support hierarchical folder structure to save
            training log.
        clsName : tuple, list
            The name of dataset labels.
        maxEpochs_1, maxEpochs_2 : int
            Maximum number of epochs in the x stage of training. Default is
            1500 and 600 respectively.
        noIncreaseEpochs : int
            Maximum number of consecutive epochs when the accuracy of the first-
            stage validation set has no relative improvement. Default is 100.
        varCheck : str
            The best value ('valInacc'/'valLoss') to check while determining 
            the best model which will be used for parameter initialization in
            the second stage of model training. Default is 'valInacc'.
            
        Returns
        -------
        Return train, validation and test results dict.
        {
            'train' : {'preds': Tensor, 'target': Tensor, 'acc': Tensor},
            'test'  : ...
            'val'   : ...,
        }
        '''
        # reset parameters of nn.Moudle and lr_scheduler
        self.net.load_state_dict(self.expDetails['origNetParam'])

        # create loss function
        self.lossFn = getattr(nn, self.lossFnType)(**self.lossFnArgs)

        # create optimizer
        self.optimizer = getattr(optim, self.optimizerType) \
            (self.net.parameters(), lr=self.lr, **self.optimArgs)
        
        # create lr_scheduler
        self.lrSch = getattr(optim.lr_scheduler, self.lrSchType) \
            (self.optimizer, **self.lrSchArgs) if self.lrSchType else None

        # check the best model
        bestVar = float('inf')
        bestNetParam = deepcopy(self.net.state_dict())
        bestOptimParam = deepcopy(self.optimizer.state_dict())

        checkList = ['valInacc', 'valLoss']
        if varCheck not in checkList:
            raise ValueError(f'Parameter `varCheck` only supports: {checkList}, '
                             f'but got {varCheck}')

        # create log writer
        logDir = os.path.abspath(logDir)
        writer = SummaryWriter(logDir)
        loger = Logger(logDir, path=os.path.join(logDir, 'running.log'), mode='a',
                       flevel='INFO', clevel=self.verbose)
        
        # initialize dataloader
        trainLoader = self._data_loader(*trainData, batchSize=self.batchSize)
        valLoader = self._data_loader(*valData, batchSize=self.batchSize)
        testLoader = self._data_loader(*testData, batchSize=self.batchSize)
        ncls = len(clsName)
        
        # start the training
        self.timer.start()
        loger.info(f'[Training...] - [{self.timer.ctime()}]')

        stopMon = ComposeStopCriteria({'Or': {
            'cri1': {'MaxEpoch': {
                'maxEpochs': maxEpochs_1, 'varName': 'epoch'
            }},
            'cri2': {'NoDecrease': {
                'numEpochs': noIncreaseEpochs, 'varName': 'valLoss'
            }}
        }})
        self.expDetails['fit'] = {
            'type': 'fit_with_val',
            'varCheck': varCheck,
            'stopMon_1': str(stopMon)
        }
        monitors = {'epoch': 0, 'valLoss': float('inf'), 'valInacc': 1,
                    'trainLoss': float('inf'), 'globalEpoch': 0}
        earlyStopReached, doStop = False, False
        
        while not doStop: 
            
            # train one epoch
            self._run_one_epoch(trainLoader)

            # evaluate the training and validation accuracy
            trainPreds, trainTarget, trainLoss = self._predict(trainLoader)
            trainAcc = self.get_acc(trainPreds, trainTarget, ncls)

            valPreds, valTarget, valLoss = self._predict(valLoader)
            valAcc = self.get_acc(valPreds, valTarget, ncls)
            monitors['valInacc'] = 1 - valAcc
            monitors['valLoss'] = valLoss

            # store loss and acc
            writer.add_scalars('train', {'loss': trainLoss, 'acc': trainAcc}, 
                                monitors['globalEpoch'])
            writer.add_scalars('val', {'loss': valLoss, 'acc': valAcc}, 
                                monitors['globalEpoch'])
            # print the epoch info
            loger.info(f'-->Epoch : {monitors["epoch"] + 1}')
            loger.info(f'  \u21b3train Loss/Acc = {trainLoss:.4f}/{trainAcc:.4f}'
                        f' | val Loss/Acc = {valLoss:.4f}/{valAcc:.4f}')

            # select best model on Stage 1
            if monitors[varCheck] <= bestVar:
                bestVar = monitors[varCheck]
                bestNetParam = deepcopy(self.net.state_dict())
                bestOptimParam = deepcopy(self.optimizer.state_dict())
                monitors['trainLoss'] = trainLoss

            monitors['epoch'] += 1
            monitors['globalEpoch'] += 1

            # check if to stop training
            if stopMon(monitors):
                # check whether to enter the second stage of training
                if not earlyStopReached:
                    earlyStopReached = True
                    loger.info('[Early Stopping Reached] -> Continue '
                                    'training on full set.')
                    # load the best state
                    self.net.load_state_dict(bestNetParam)
                    self.optimizer.load_state_dict(bestOptimParam)

                    # Combine the train and val dataset
                    trainLoader = self._data_loader(*trainData, *valData,
                                                    batchSize=self.batchSize)

                    # update stop monitor and epoch
                    stopMon = ComposeStopCriteria({'Or': {
                        'cri1': {'MaxEpoch': {
                            'maxEpochs': maxEpochs_2, 'varName': 'epoch'
                        }},
                        'cri2': {'Smaller': {
                            'var': monitors['trainLoss'], 'varName': 'valLoss'
                        }}
                    }})
                    self.expDetails['fit']['stopMon_2'] = str(stopMon)
                    monitors['epoch'] = 0
                else:
                    bestNetParam = deepcopy(self.net.state_dict())
                    doStop = True
        
        writer.close()

        # report the checkpoint time of end and compute cost time
        loger.info(f'[Train Finish] - [{self.timer.ctime()}]')
        h, m, s = self.timer.stop()
        loger.info(f'Cost Time = {h}H:{m}M:{s:.2f}S')

        # load the best model and evaulate this model in testData
        self.net.load_state_dict(bestNetParam)

        results = {}
        trainPreds, trainTarget, trainLoss = self._predict(trainLoader)
        trainAcc = self.get_acc(trainPreds, trainTarget, ncls)
        results['train'] = {
            'preds': trainPreds, 'target': trainTarget, 'acc': trainAcc
        }
        valPreds, valTarget, valLoss = self._predict(valLoader)
        valAcc = self.get_acc(valPreds, valTarget, ncls)
        results['val'] = {
            'preds': valPreds, 'target': valTarget, 'acc': valAcc
        }
        testPreds, testTarget, testLoss = self._predict(testLoader)
        testAcc = self.get_acc(testPreds, testTarget, ncls)
        results['test'] = {
            'preds': testPreds, 'target': testTarget, 'acc': testAcc
        }

        loger.info(f'Loss: train={trainLoss:.4f} | val={valLoss:.4f} | '
                   f'test={testLoss:.4f}')
        loger.info(f'Acc:  train={trainAcc:.4f} | val={valAcc:.4f} | '
                   f'test={testAcc:.4f}')

        # save the experiment details
        self.expDetails['result'] = results
        self.expDetails['bestNetParam'] = bestNetParam

        # store the training details
        with open(os.path.join(logDir, f'train.pkl'), 'wb') as f:
            pickle.dump(self.expDetails, f)

        # store the best net model parameters
        modelPath = os.path.join(logDir, f'train_checkpoint_best.pth')
        torch.save(bestNetParam, modelPath)
        
        return results

    def _run_one_epoch(
        self,
        trainLoader : DataLoader
    ) -> None:
        '''Run one epoch to train net.

        Parameters
        ----------
        trainLoader : DataLoader
            DataLoader used for training.
        '''
        # set the network in training mode
        self.net.train()

        # iterater over all the data
        with torch.enable_grad():
            for idx, (data, label) in enumerate(trainLoader):
                data, label = data.to(self.device), label.to(self.device)
                out = self.net(data)
                loss = self.lossFn(out, label)
                loss.backward()
                # gradient accumulation
                if((idx + 1) % self.gradAcc == 0):
                    # 1 - update parameters
                    self.optimizer.step()
                    # 2 - zero the parameter gradients
                    self.optimizer.zero_grad()
            # update lr
            # Note: Learning rate scheduling should be applied after optimizer’s update
            if self.lrSch:
                self.lrSch.step()

    def _predict(
        self,
        dataLoader : DataLoader,
    ) -> Tuple[Tensor, Tensor, float]:
        '''Predict the class of the input data.

        Parameters
        ----------
        dataLoader : DataLoader
            Dataset used for prediction.

        Returns
        -------
        preds : Tensor
            Predicted labels, as returned by a classifier.
        target : Tensor
            Ground truth (correct) labels.
        loss : float
            Average loss.
        '''
        preds, target = [], []
        lossSum = 0

        # set the network in the eval mode
        self.net.eval()

        # iterate over all the data
        with torch.no_grad():
            for data, label in dataLoader:
                data, label = data.to(self.device), label.to(self.device)
                out = self.net(data)
                lossSum += self.lossFn(out, label).detach().item()
                # convert the output of soft-max to class label
                # save preds and actual label
                preds.append(torch.argmax(out, dim=1).detach().cpu())
                target.append(label.cpu())
        
        return torch.cat(preds), torch.cat(target), lossSum / len(dataLoader)
    
    def get_acc(self, preds : Tensor, target : Tensor, ncls : int) -> Tensor:
        '''Easy for program to caculate the accuarcy.
        '''
        return accuracy(preds, target, 'multiclass', num_classes=ncls)

    def set_seed(self, seed : int = DPEEG_SEED) -> None:
        '''Sets the seed for generating random numbers for cpu and gpu.
        '''
        torch.manual_seed(seed)
        if self.device != torch.device('cpu'):
            torch.cuda.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
        self.loger.info(f'Set all random seed = {seed}')

    def get_device(self, nGPU : int = 0) -> torch.device:
        '''Get the device for training and testing.

        Parameters
        ----------
        nGPU : int
            GPU number to train on. Default is 0.
        '''
        if not torch.cuda.is_available():
            self.loger.info('GPU is not avaiable and the CPU will be used')
            dev = torch.device('cpu')
        else:
            if nGPU > torch.cuda.device_count() - 1:
                raise ValueError(f'GPU: {nGPU} does not exit.')
            dev = torch.device(f'cuda:{nGPU}')
            self.loger.info(f'Network will be trained in "cuda:{nGPU} ' +
                            f'({torch.cuda.get_device_name(dev)})"')
        return dev
    
    def _data_loader(
        self, 
        *datasets,
        batchSize : int = 256,
    ) -> DataLoader:
        '''Wrap multiple sets of data and labels and return DataLoader.

        Parameters
        ----------
        dataset : evey two sequence of indexables with same length / shape[0]
            Allowed inputs are lists and tuple of tensor or ndarray.
        '''
        nDataset = len(datasets)
        if nDataset % 2:
            raise ValueError('One data set corresponds to one label set, but the '
                             f'total number of data got is {nDataset}.')
        if nDataset == 0:
            raise ValueError('At least one dataset required ad input.')

        # dataset wrapping tensors
        td = []
        for i in range(0, nDataset, 2):
            td.append(TensorDataset(*to_tensor(datasets[i], datasets[i + 1])))
        td = ConcatDataset(td)

        return DataLoader(td, batchSize, True, pin_memory=True)

    def __repr__(self) -> str:
        '''Trainer details.
        '''
        s = '[Network architecture]:\n' + self.netArch
        s += f'[Loss function]:\t{self.lossFnType}\n'
        if self.lossFnArgs:
            s += f'[LossFn Args]:\t{self.lrSchArgs}\n'
        s += f'[Optimizer]:\t{self.optimizerType}\n'
        s += f'[Learning rate]:\t{self.lr}\n'
        if self.optimArgs:
            s += f'[Optim Args]:\t{self.optimArgs}\n'
        if self.lrSchType:
            s += f'[Lr scheduler]:\t{self.lrSchType}\n'
            if self.lrSchArgs:
                s += f'[LrSch Args]:\t{self.lrSchArgs}\n'
        s += f'[Grad Acc]:\t{self.gradAcc}\n'
        return s



#!/usr/bin/env python
# coding: utf-8

"""
    Basic method of `transforms.py` file.
    
    @Author  : SheepTAO
    @Time    : 2023-09-14
"""


import os
import torch
import numpy as np
from numpy import ndarray
from torch import Tensor
from typing import Optional, Union, Tuple, List

from ..utils import loger, verbose, DPEEG_SEED


@verbose
def split_train_test(
    *arrs, 
    testSize : float = .2, 
    seed : int = DPEEG_SEED, 
    sample : Optional[List[int]] = None, 
    verbose : Optional[Union[int, str]] = None
) -> list:
    '''Split an dataset into training and testing sets. The axis along which
    to split is 0.

    Parameters
    ----------
    *arrs : sequence of indexables with same length / shape[0]
        Allowed inputs are lists and numpy arrays.
    testSize : float
        The proportion of the test set. Default is 0.2. If index is not None,
        testSize will be ignored. Default use stratified fashion and the last
        arr serves as the class labels.
    seed : int
        Random seed when splitting. Default is DPEEG_SEED.
    sample : list of int, optional
        A list of integers, the entries indicate which data were selected
        as the test set. If None, testSize will be used. Default is None.
    '''
    nArrs = len(arrs)
    if nArrs == 0:
        raise ValueError('At least one array required as input.')

    arrList = [np.array(arr) for arr in arrs]
    lengths = [len(arr) for arr in arrList]
    uniques = np.unique(lengths)
    if len(uniques) > 1:
        raise ValueError(
            'Found input variables with inconsistent numbers of samples: %r'
            % [int(l) for l in lengths]
        )
        
    if sample:
        smparr = np.array(sample)
        if np.unique(smparr).size != smparr.size:
            raise IndexError(
                f'Found repeated sampling of test set: {smparr.tolist()}.'
            )
        length = lengths[0]
        if smparr.size >= uniques:
            raise IndexError(f'The number of samples (={len(smparr)}) in the '
                             'test set cannot exceed the total number of data '
                             f'sets (={length}).'
            )

        testSample, trainSample = smparr, np.setdiff1d(np.arange(length), smparr)
        res = []
        for arr in arrList:
            res.extend([arr[trainSample], arr[testSample]])
        return res
    else:
        from sklearn.model_selection import train_test_split
        return train_test_split(*arrList, test_size=testSize, random_state=seed,
                                stratify=arrList[-1])


@verbose
def to_tensor(
    data: Union[Tensor, ndarray],
    label: Union[Tensor, ndarray],
    verbose : Optional[Union[int, str]] = None
) -> Tuple[Tensor, Tensor]:
    '''Convert the numpy data and label into trainable Tensor format.
    '''
    dataT = torch.from_numpy(data).float() \
        if isinstance(data, ndarray) else data.float()
    labelT = torch.from_numpy(label).long() \
        if isinstance(label, ndarray) else label.long()
    if dataT.size(0) != labelT.size(0):
        loger.warning('Data and label do not match in the first dimension: '
                      f'{dataT.size(0)} and {labelT.size(0)} respectively.')
    return dataT, labelT


@verbose
def slide_win(
    data : ndarray,
    win : int, 
    overlap : int = 0,
    label : Optional[ndarray] = None,
    verbose : Optional[Union[int, str]] = None,
) -> Union[Tuple[ndarray, ndarray], ndarray]:
    '''This transform is only splits the time series (dim = -1) through the sliding 
    window operation on the original dataset. If the time axis is not divisible by 
    the sliding window, the last remaining time data will be discarded.

    Parameters
    ----------
    data : array of float, shape (..., times)
        The data to split.
    win : int
        The size of the sliding window.
    overlap : int
        The amount of overlap between adjacent sliding windows. Default is 0.
    label : ndarray, optional
        The label of the data. If not None, label will update with sliding window.
        Default is None.

    Returns
    -------
    If label is not None, (data, label) will be returned, or data will be returned.
    '''
    if win < 0 or overlap < 0:
        raise ValueError('win and overlap only allow positive numbers, '
                         f'but got {win} and {overlap} respectively.')
    if overlap >= win:
        raise ValueError(f'overlap={overlap} should be less than win={win}.')

    loger.info(f'Sliding window with win:{win} and overlap:{overlap}.')
    winSize = win - overlap
    sldNum = data.shape[-1] // winSize
    if not sldNum:
        loger.warning('The sliding window is larger than the data to be split.')
    dataList = [data[..., : winSize]]
    for i in range(1, sldNum):
        sld = data[..., i*winSize-overlap :(i+1)*winSize-overlap]
        dataList.append(sld)

    data = np.concatenate(dataList)
    if isinstance(label, ndarray):
        label = np.repeat(label, sldNum)
        return data, label
    return data


@verbose
def save(
    folder : str,
    input : dict,
    verbose : Optional[Union[int, str]] = None,
) -> None:
    '''Save transformed dataset to a binary file in NumPy `.npy` format.

    Parameters
    ----------
    folder : str
        Folder name to save transformed data.
    input : dict
        Data to be saved.
    '''
    folder = os.path.abspath(folder)
    os.makedirs(folder, exist_ok=True)
    if os.listdir(folder):
        raise FileExistsError(f'\'{folder}\' is not a empty folder.')
    
    loger.info(f'Transformed data will be saved in: \'{folder}\'')
    for sub, data in input.items():
        loger.info(f'Save transformed data of sub_{sub}.')
        for mode in ['train', 'test']:
            fileName = os.path.join(folder, f'{sub}_{mode}')
            np.save(fileName + '_data', data[mode][0])
            np.save(fileName + '_label', data[mode][1])


@verbose
def load(
    folder : str,
    subjects : Optional[List[int]] = None,
    mode : str = 'all',
    verbose : Optional[Union[int, str]] = None,
) -> dict:
    '''Load saved transformed dataset from folder.

    Parameters
    ----------
    folder : str
        Folder name where transformed data is saved.
    subjects : list of int, optional
        List of subject number. If None, all subjects will be loaded. Default is None.
    mode : str
        Mode to load data. If 'all', both train set and test set will be loaded. If 
        'train', only train set will be loaded. 'test' is the same. Default is 'all'.
    '''
    path = os.path.abspath(folder)
    loger.info(f'Loading dataset from \'{path}\'')

    pathList = os.listdir(path)
    subList = list(set([int(p.split('_')[0]) for p in pathList]))
    subList.sort()
    if subjects:
        intersection = set(subjects) & set(subList)
        exclude = set(subjects) - set(subList)
        if exclude:
            loger.warning(f'Could not find subjects: {exclude}, ' +
                          f'only load subjects: {intersection}')
        subList = list(intersection)

    dataset = {}
    for sub in subList:
        loger.info(f'Loading subject {sub}')
        fileName = os.path.join(path, str(sub))
        if mode.lower() == 'train' or mode.lower() == 'all':
            data = np.load(fileName + '_train_data.npy', allow_pickle=True)
            label = np.load(fileName + '_train_label.npy', allow_pickle=True)
            dataset.setdefault(sub, {})['train'] = [data, label]
        if mode.lower() == 'test' or mode.lower() == 'all':
            data = np.load(fileName + '_test_data.npy', allow_pickle=True)
            label = np.load(fileName + '_test_label.npy', allow_pickle=True)
            dataset.setdefault(sub, {})['test'] = [data, label]
        if mode not in ['train', 'test', 'all']:
            raise KeyError(f'Only support `all`, `train` and `test` mode.')
    loger.info('Load dataset done.')
    return dataset



#!/usr/bin/env python
# coding: utf-8

"""
    This module is used to evaluate the metrics of the model and save.

    @author: SheepTAO
    @data: 2023-4-13
"""


import torch
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Union
from torchmetrics.functional.classification.confusion_matrix import confusion_matrix


def save_cm_img(
    preds: torch.Tensor,
    target: torch.Tensor,
    clsName : Union[list, tuple],
    figPath : str,
) -> None:
    '''Calculate and save the corresponding confusion matrix figure to the given
    img path.

    Parameters
    ----------
    preds : Tensor
        Predicted labels, as returned by a classifier.
    target : Tensor
        Ground truth (correct) labels.
    clsName : list, tuple
        The name of dataset labels.
    figPath : str
        Path to store the figure.
    '''
    cm = confusion_matrix(preds, target, 'multiclass', num_classes=len(clsName))
    fig = sns.heatmap(cm / cm.sum(1, keepdim=True), annot=True, cmap='Blues',
                      xticklabels=clsName, yticklabels=clsName) # type: ignore
    plt.savefig(figPath)
    # !!! Clear the current figure
    plt.clf()



#!/usr/bin/env python
# coding: utf-8

"""
    @Desc    : A simple timer to record the running time of the program.
    @Author  : SheepTAO
    @Time    : 2023-07-24
"""


import time
from typing import Tuple


class Timer:
    '''Timing the running time of the program.

    Examples
    --------
    If you want to start a default timer and then calculate the program interval,
    you can do like:
    >>> timer = Timer()
    >>> do_something()
    >>> h, m, s = timer.stop()
    >>> h, m, s
        1, 2, 13.435
        
    And you can update the start time of the default timer, just do like:
    
    >>> timer.start()

    You can also register a specified timer and update the start time of it.

    >>> timer.start('new')
    >>> h, m, s = timer.stop('new')
    '''
    def __init__(self) -> None:
        '''Timing the running time of the program.
        Initialization will automatically start a default timer (dpeeg_timer).
        '''
        self._beg = {}
        self.start('dpeeg_timer')

    def start(self, name : str = 'dpeeg_timer') -> str:
        '''Update specified timer or register a timer if it does not exist.
        '''
        self._beg[name] = time.time()
        return self.ctime()

    def stop(self, name : str = 'dpeeg_timer') -> Tuple[int, int, float]:
        '''Get the hours, minutes, and seconds of a specific timer.
        '''
        sec = self.stop_precise(name)
        m, s = divmod(sec, 60)
        h, m = divmod(m, 60)

        return int(h), int(m), s

    def stop_precise(self, name : str = 'dpeeg_timer') -> float:
        '''Get the precise time for a specific timer.
        '''
        if name not in self._beg.keys():
            raise RuntimeError(f'Timer `{name}` has not been registered yet.')
        ct = time.time()
        return ct - self._beg[name]

    @staticmethod
    def ctime() -> str:
        return time.ctime()



from dpeeg import __version__
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='dpeeg',
    version=__version__,
    author='SheepTAO',
    author_email='sheeptao@outlook.com',
    license='MIT',
    description='Deep learning with EEG',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/SheepTAO/dpeeg',
    packages=find_packages(),
    platforms=['all'],
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',
        
        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ],
)


#!/usr/bin/env python
# coding: utf-8

"""
    @Desc    : A Logger to implement terminal output and log file writing.
    @Author  : SheepTAO
    @Time    : 2023-07-20
"""


import os, sys, logging
from typing import Union, Optional


class Logger:
    def __init__(
        self,
        loger : str = 'dpeeg_root',
        path : Optional[str] = None,
        mode : str = 'a',
        clevel : Union[int, str] = logging.INFO,
        flevel : Optional[Union[int, str]] = None,
    ) -> None:
        '''Logging hooks for terminals and file streams.

        Can handle terminal printing and file writing in three situations:
        1.Terminal print;
        2.Terminal and file streams write simultaneously;
        3.Ordinary file write.

        Parameters
        ----------
        loger : str
            Specify the logger name, creating it if necessary. If no name is specified, 
            return the root logger. Default is dpeeg_root.
        path : str, optional
            The path of log file. Default is None.
        mode : str
            The write mode of the log file. Default is 'a'.
        clevel : int, str
            The log level of console. Default is DEBUG.
        flevel : int, str, optional
            The log level of filehandler. If it is None, the log file will not be written
            and parameter path will be ignored. Default is None.
        '''

        self._logger = logging.getLogger(loger)
        self._logger.propagate = False
        self._logger.setLevel(logging.DEBUG)

        self._sh = logging.StreamHandler()
        shfmt = logging.Formatter('%(message)s')
        self._sh.setFormatter(shfmt)
        self._sh.setLevel(clevel)
        self._logger.addHandler(self._sh)
        if flevel != None:
            assert path != None, 'path cannot be empty.'
            self._fh = logging.FileHandler(os.path.abspath(path), mode)
            fhfmt = logging.Formatter('[%(asctime)s] [%(levelname)8s]: %(message)s', 
                                    datefmt='%Y-%m-%d %H:%M:%S')
            self._fh.setFormatter(fhfmt)
            self._fh.setLevel(flevel)
            self._logger.addHandler(self._fh)

        sys.excepthook = self.handle_exception

    def _update_sh_level(self, level : Union[int, str]):
        self._sh.setLevel(level)

    def _get_sh_level(self) -> str:
        return logging.getLevelName(self._sh.level)

    def debug(self, message : str):
        self._logger.debug(f'[DEBUG]: {message}')

    def info(self, message : str):
        self._logger.info(message)

    def warning(self, message : str):
        self._logger.warning(f'[WARNING]: {message}')

    def error(self, message : str):
        self._logger.error(f'[ERROR]: {message}')

    def critical(self, message : str):
        self._logger.critical(f'[CRITICAL]: {message}')

    def handle_exception(self, exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
        self._logger.critical("Uncaught exception>>>", exc_info=(exc_type, exc_value, exc_traceback))
